from typing import Dict, List, Union, Optional

from typeguard import typechecked

from . import _kv_engine


@typechecked
def create_engine(
    *,
    engine: str,
    reason: Optional[str] = "Task is creating a new secrets engine in Hashicorp Vault",
    vault_name: Optional[str] = None,
    kv_version: Optional[int] = 2,
) -> None:
    """ Create a new Key-Value secrets engine in a Hashicorp Vault engine.

    :param engine: The name of the secret engine.
    :param reason: Optional. The reason for accessing the secret in Hashicorp vault.
        Defaults to: `Task is creating a new secrets engine in Hashicorp Vault`.
    :param vault_name: Optional. The name of the Hashicorp Vault configured in the PPA Appliance. Use this to target a
        specific Hashicorp Vault if your PPA Appliance has more than one configured. If more than one is configured and
        you do not provide a `vault_name`, the task operator will need to choose a Hashicorp Vault from a table in the
        PPA task interface.
    :param kv_version: Optional. The version of the KV secrets engine. Defaults to 2.

    :raises exceptions.EngineAlreadyExists: if an engine with the supplied name already exists.
    """
    return _kv_engine._create_engine(engine=engine, reason=reason, vault_name=vault_name, kv_version=kv_version)


@typechecked
def create_secret(
    *,
    secret: str,
    engine: str = "ppa",
    data: Dict[str, str],
    reason: Optional[str] = "Task is creating a new secret in Hashicorp Vault",
    vault_name: Optional[str] = None,
    kv_version: Optional[int] = 2,
    overwrite: Optional[bool] = False,
) -> None:
    """ Create a new secret in a Key-Value engine.

    :param secret: The name of the secret in the KV secrets engine.
    :param engine: Optional. The name of the secret engine. Defaults to "ppa",
        matching the default value in Hashicorp Vault.
    :param reason: Optional. The reason for accessing the secret in Hashicorp vault.
        Defaults to: `Task is creating a new secret in Hashicorp Vault`.
    :param vault_name: Optional. The name of the Hashicorp Vault configured in the PPA Appliance. Use this to target a
        specific Hashicorp Vault if your PPA Appliance has more than one configured. If more than one is configured and
        you do not provide a `vault_name`, the task operator will need to choose a Hashicorp Vault from a table in the
        PPA task interface.
    :param kv_version: Optional. The version of the KV secrets engine. Defaults to 2.
    :param overwrite: Optional. Defaults to `False`. Set to `True` to overwrite any existing secret
        with the supplied name.


    :raises exceptions.SecretAlreadyExists: if a secret with the supplied name already
        exists, and overwrite is not set to `True`.
    :raises exceptions.InvalidEngine: if the secret engine does not exist in the Hashicorp Vault.
    """
    return _kv_engine._create_secret(
        secret=secret,
        engine=engine,
        data=data,
        reason=reason,
        vault_name=vault_name,
        kv_version=kv_version,
        overwrite=overwrite,
    )


@typechecked
def read_secret(
    *,
    secret: str,
    keys: Union[List[str], None] = None,
    reason: Optional[str] = "Task is accessing keys from Hashicorp Vault",
    engine: str = "ppa",
    vault_name: Optional[str] = None,
    kv_version: Optional[int] = 2,
):
    """ Read keys from a KV Engine secret.

    :param secret: The name of the secret in the KV secrets engine.
    :param keys: A list of strings containing the keys to return from the secret. If None, all the keys are returned.
    :param engine: Optional. The name of the secret engine. Defaults to "secret",
        matching the default value in Hashicorp Vault.
    :param reason: Optional. The reason for accessing the secret in Hashicorp vault.
        Defaults to: `Task is accessing keys from Hashicorp Vault`.
    :param vault_name: Optional. The name of the Hashicorp Vault configured in the PPA Appliance. Use this to target a
        specific Hashicorp Vault if your PPA Appliance has more than one configured. If more than one is configured and
        you do not provide a `vault_name`, the task operator will need to choose a Hashicorp Vault from a table in the
        PPA task interface.
    :param kv_version: Optional. The version of the KV secrets engine. Defaults to 2.


    :raises exceptions.InvalidSecret: if the secret does not exist in the KV engine, or if does
        exist but is held in an engine that uses an unsupported version.
    :raises exceptions.InvalidEngine: if the secret engine does not exist in the Hashicorp Vault.
    :raises exceptions.InvalidKey: if a supplied key does not exist in the secret.
    :raises exceptions.EmptySecret: if the supplied secret exists but is empty.
    :raises exceptions.VaultError: if a secret exists but contains no data, or if a response
        from the KV secrets engine.

    :returns: A namedtuple record containing either all available fields, or the
        fields requested with the `keys` parameter.

    .. code-block:: python

        # read the username, password, and address keys
        # from the secret called 'windows_server'.
        windows_secrets = kv_engine.read_secret(
            secret="windows_server",
            keys=["username", "password", "address"],
        )

        # access the field values with dot-notation
        windows_secrets.username
        windows_secrets.password
        windows_secrets.address

    """
    return _kv_engine._read_secret(
        secret=secret, engine=engine, keys=keys, reason=reason, vault_name=vault_name, kv_version=kv_version
    )


@typechecked
def update_secret(
    *,
    secret: str,
    engine: str = "ppa",
    data: Dict[str, str],
    reason: Optional[str] = None,
    vault_name: Optional[str] = None,
    kv_version: Optional[int] = 2,
) -> None:
    """ Update the values in a KV Engine secret. This function can be used to:

        - Replace values in a secret
        - Append new values to a secret

    :param secret: The name of the secret in the KV secrets engine.
    :param engine: Optional. The name of the secret engine. Defaults to "ppa",
        matching the default value in Hashicorp Vault.
    :param data: A dictionary of new keys and values to apply to the secret.
    :param reason: Optional. The reason for accessing the secret in Hashicorp vault.
        Defaults to: `Task is updating {x} keys in Hashicorp Vault`.
    :param vault_name: Optional. The name of the Hashicorp Vault configured in the PPA Appliance. Use this to target a
        specific Hashicorp Vault if your PPA Appliance has more than one configured. If more than one is configured and
        you do not provide a `vault_name`, the task operator will need to choose a Hashicorp Vault from a table in the
        PPA task interface.
    :param kv_version: Optional. The version of the KV secrets engine. Defaults to 2.

    :raises exceptions.InvalidEngine: if the secret engine does not exist in the Hashicorp Vault.
    :raises exceptions.VaultError: if the update to the KV secrets engine returns an unexpected
        status code.

    .. code-block:: python

        # update the username and password keys in
        # the secret called 'windows_server'.
        kv_engine.update_secret(
            secret="windows_server",
            engine="kv",
            data={
                "username": "my.username",
                "password": "my.password",
            }
        )

    """
    return _kv_engine._update_secret(
        secret=secret, engine=engine, data=data, reason=reason, vault_name=vault_name, kv_version=kv_version
    )
