"""Key-Value Engine

This module contains actions for:

- Creating Secrets Engines
- Creating Secrets
- Reading Secrets
- Updating Secrets

!!! question "Supported Versions"

    This plugin supports Hashicorp Vault Key-Value versions 1 & 2.

!!! danger "Remember"

    You must have a Hashicorp Vault provisioned as a _Vault_ inside PPA to use this plugin.
"""
import functools
import logging

from typing import Dict, List, Optional, Any, Union, Callable, TypedDict

from ppa import ui
from hashicorp_vault import kv_engine, exceptions
from hashicorp_vault._kv_engine import vaults

from . import _markdown_templates


logger = logging.getLogger(__name__)


class VaultKey(TypedDict):
    """
- Describes a key inside a Hashicorp Vault secret

- Used by the [read_secret_keys](../hashicorp_vault.key_value/#read_secret_keys) action to both retrieve & create missing secret keys

!!! warning "Required Keys"

    __name__: The name of the key (required).

!!! info "Optional Keys"

    __format__:  Either `string` or `file` (defaults to `string`).

    __sensitive__: Either `true` or `false` (defaults to `false`).

!!! example "Sensitive Password"

    ```YAML linenums="1" hl_lines="3 4 5 6 7 8"
      - hashicorp_vault.key_value.read_secret_keys:
          secret: active_directory
          keys:
            - name: address
            - name: domain
            - name: username
            - name: password
              sensitive: true
          reason: Getting Active Directory details
          create_missing: true
        save: domain_controller
    ```

!!! example "Sensitive Public Key File"

    ```YAML linenums="1" hl_lines="3 4 5 6 7 8 9 10 11"
      - hashicorp_vault.key_value.read_secret_keys:
          secret: ssh
          keys:
            - name: address
            - name: port
            - name: username
            - name: private_key
              format: file
              sensitive: true
            - name: key_passphrase
              sensitive: true
          reason: Getting SSH Server Details
          create_missing: true
        save: ssh_server
    ```
"""

    name: str
    format: Optional[str] = "string"
    sensitive: Optional[bool] = False


def _process_key(key: Dict[str, Any]) -> VaultKey:
    supported_formats = {"file", "string"}
    default_format = "string"  # Default value
    default_sensitive = False  # Default value

    try:
        name = key["name"]
    except KeyError:
        message = "A vault key was supplied without a name."
        ui.output_error(message)
        raise ValueError(message)

    # Flake8 will complain if these variables don't have values before the try/excepts
    format = None
    sensitive = None

    try:
        format = key.get("format", default_format)
        assert format in supported_formats
    except AssertionError:
        message = "Vault key '{}' contains unsupported format '{}'. Supported formats are '{}'.".format(
            name, format, ", ".join(supported_formats)
        )
        ui.output_error(message)
        raise ValueError(message)

    try:
        sensitive = key.get("sensitive", default_sensitive)
        assert isinstance(sensitive, bool)
    except AssertionError:
        message = f"The 'sensitive' value for vault key '{name}' must be true or false. Received '{sensitive}'."
        ui.output_error(message)
        raise ValueError(message)

    return VaultKey(name=name, sensitive=sensitive, format=format)


def _catch_no_vaults(func: Callable):
    @functools.wraps(func)
    def _vault_request(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except vaults.NoVaultsFound as e:
            ui.output_error(e.args[0])
            raise e

    return _vault_request


@_catch_no_vaults
def create_engine(
    engine: str = "ppa", reason: Optional[str] = None, vault_name: Optional[str] = None, kv_version: Optional[int] = 2,
) -> None:
    """Create a new secrets engine in a Hashicorp Vault.

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __engine:__ the name of the secrets engine (defaults to `ppa`)

    - __reason:__ task interface audit message (defaults to a message with the engine name)

    - __vault_name:__ the Hashicorp Vault name in PPA (only necessary if more than one vault is configured)

    - __kv_version:__ the version of the key value engine (defaults to `2`)

??? success "Output"
    Nothing is outputted by this action.

??? example

    Creating a new v2 engine called `production_secrets`.

    ```YAML linenums="1"
    - hashicorp_vault.key_value.create_engine:
        engine: production_secrets
    ```
    """
    return kv_engine.create_engine(
        engine=engine,
        reason=reason or f"Task is creating new secrets engine '{engine}' in Hashicorp Vault",
        vault_name=vault_name,
        kv_version=kv_version,
    )


@_catch_no_vaults
def create_secret(
    secret: str,
    data: Dict[str, str],
    engine: str = "ppa",
    reason: Optional[str] = None,
    vault_name: Optional[str] = None,
    kv_version: Optional[int] = 2,
    overwrite: Optional[bool] = False,
) -> None:
    """Create a new secret in a Key-Value engine.

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __secret:__ the name of the new secret

    - __data:__ a dictionary containing keys & values to create in the secret

    - __engine:__ the name of the secrets engine (defaults to `ppa`)

    - __reason:__ task interface audit message (defaults to a message with the secret & engine names)

    - __vault_name:__ the Hashicorp Vault name in PPA (only necessary if more than one vault is configured)

    - __kv_version:__ the version of the key value engine (defaults to `2`)

    - __overwrite:__ set to `true` to overwrite any existing secret with the same name (defaults to `false`)

??? success "Output"
    Nothing is outputted by this action.

??? example

    Creating a new secret with a username & password in the default `ppa` v2 engine.
    {% raw %}
    ```YAML linenums="1"
    - hashicorp_vault.key_value.create_secret:
        secret: my_new_secret
        data:
          username: "{{ secret_username }}"
          password: "{{ secret_password }}"
    ```
    {% endraw %}
    """
    return kv_engine.create_secret(
        secret=secret,
        engine=engine,
        data=data,
        reason=reason or f"Task is creating new secret '{secret}' in Hashicorp Vault secrets engine '{engine}'",
        vault_name=vault_name,
        kv_version=kv_version,
        overwrite=overwrite,
    )


@_catch_no_vaults
def update_secret(
    secret: str,
    data: Dict[str, Any],
    engine: Optional[str] = "ppa",
    reason: Optional[str] = None,
    vault_name: Optional[str] = None,
    kv_version: Optional[int] = 2,
) -> None:
    """Update the values of a secret in a Key-Value engine.

This action can be used to:

- Replace values in a secret
- Append new values to a secret

This action cannot _create_ a new secret, use [create_secret](#create_secret) to do this.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __engine:__ the name of the secrets engine (defaults to `ppa`)

    - __secret:__ the name of the secret in the secrets engine

    - __data:__ a dictionary containing keys & values to update in the secret

    - __reason:__ task interface audit message (defaults to a message with the secret name)

    - __vault_name:__ the Hashicorp Vault name in PPA (only necessary if more than one vault is configured)

    - __kv_version:__ the version of the key value engine (defaults to `2`)

??? success "Output"
    Nothing is outputted by this action.

??? example

    Updating the _service_now_ secret's _access_token_ value in the _production_ engine.

    ```YAML linenums="1"
    - hashicorp_vault.key_value.update_secret:
        secret: service_now
        engine: production
      load:
        data:
          access_token: new_access_token
    ```
    """
    return kv_engine.update_secret(
        engine=engine,
        secret=secret,
        data=data,
        reason=reason or f"Task is updating {len(data)} values for secret '{secret}' in Hashicorp Vault",
        vault_name=vault_name,
        kv_version=kv_version,
    )


@_catch_no_vaults
def read_secret(
    secret: str,
    engine: Optional[str] = "ppa",
    reason: Optional[str] = None,
    vault_name: Optional[str] = None,
    kv_version: Optional[int] = 2,
):
    """Read all values from a secret in a Key-Value engine.

If you know which keys you need, consider using [read_secret_keys](#read_secret_keys) instead.

Unlike [read_secret_keys](#read_secret_keys), this action will not:

- Start a config wizard if configuration is missing
- Perform any checks on the keys read from the secret

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __engine:__ the name of the secrets engine (defaults to `ppa`)

    - __secret:__ the name of the secret in the secrets engine

    - __reason:__ task interface audit message (defaults to a message with the secret name)

    - __vault_name:__ the Hashicorp Vault name in PPA (only necessary if more than one vault is configured)

    - __kv_version:__ the version of the key value engine (defaults to `2`)

??? success "Output"

    A dictionary containing all keys & values from the secret.

??? example

    - Getting all keys from the _active_directory_ secret in the _production_ engine

    - Saving them as a new variable called `active_directory_secrets`

    ```YAML linenums="1"
    - hashicorp_vault.key_value.read_secret:
        secret: active_directory
        engine: production
      save: active_directory_secrets
    ```
"""
    return kv_engine.read_secret(
        engine=engine,
        secret=secret,
        reason=reason or f"Task is reading all values from secret '{secret}'",
        vault_name=vault_name,
        kv_version=kv_version,
    )._asdict()


@_catch_no_vaults
def read_secret_keys(
    secret: str,
    keys: List[VaultKey],
    engine: Optional[str] = "ppa",
    reason: Optional[str] = None,
    vault_name: Optional[str] = None,
    kv_version: Optional[int] = 2,
    create_missing: Optional[bool] = False,
) -> Dict[str, Any]:
    """Read specific values from a secret in a Key-Value engine.

!!! question "Minimum Plugin Version: 5.0.0"

!!! hint "Vault Setup Wizard"

    PPA can help the Task Operator create any missing engine, secret, or keys supplied to this action.

    If __create_missing__ is `true` & required configuration is missing, the setup wizard will run.

    The __create_missing__ input defaults to `false` for security reasons.

??? info "Input"

    - __engine:__ the name of the secrets engine (defaults to `ppa`)

    - __secret:__ the name of the secret in the secrets engine

    - __keys:__ a list of [VaultKey](../inputs_outputs/#vaultkey) dictionaries defining what to retrieve from the secret

    - __reason:__ task interface audit message (defaults to a message with the secret name & number of keys)

    - __vault_name:__ the Hashicorp Vault name in PPA (only necessary if more than one vault is configured)

    - __kv_version:__ the version of the key value engine (defaults to `2`)

    - __create_missing:__ if `true` the Vault setup wizard will start if any configuration is missing (defaults to `false`)

??? success "Output"

    A dictionary containing the keys & values from the secret.

??? example "Active Directory"

    Retrieving 4 keys from the `active_directory` secret & saving them as `domain_controller`.

    If any keys are missing:

    - The Vault Setup Wizard will allow the user to supply them

    - The `password` key is marked as sensitive & will be obfuscated when supplied

    ```YAML linenums="1"
      - hashicorp_vault.key_value.read_secret_keys:
          secret: active_directory
          keys:
            - name: address
            - name: domain
            - name: username
            - name: password
              sensitive: true
          reason: Getting Active Directory details
          create_missing: true
        save: domain_controller
    ```

??? example "SSH With Private Key"

    Retrieving 5 keys from the `ssh` secret & saving them as `ssh_server`.

    If any keys are missing:

    - The Vault Setup Wizard will allow the user to supply them

    - The `private_key` is marked as both sensitive & as a file, & any uploaded file contents will be obfuscated

    - The `key_passphrase` is marked as sensitive & will be obfuscated when supplied

    ```YAML linenums="1"
      - hashicorp_vault.key_value.read_secret_keys:
          secret: ssh
          keys:
            - name: address
            - name: port
            - name: username
            - name: private_key
              format: file
              sensitive: true
            - name: key_passphrase
              sensitive: true
          reason: Getting SSH Server Details
          create_missing: true
        save: ssh_server
    ```
"""
    processed_keys = [_process_key(k) for k in keys]
    if rogue_keys := [k for k in [key["name"] for key in keys] if k not in [key["name"] for key in processed_keys]]:
        raise KeyError(
            "The following supplied keys could not be processed into sensitive/insensitive string or file keys: '{}'.".format(
                ", ".join(rogue_keys)
            )
        )

    def _format_keys(keys: List[VaultKey]) -> str:
        return "\n".join([f"- {key['name']}" for key in keys])

    def _read(
        reason: str, key_list: Union[List[VaultKey], None] = processed_keys  # noqa (ignore default mutable value)
    ) -> dict:  # If we need to find which keys are missing, the keys argument is supplied as None.
        return kv_engine.read_secret(
            engine=engine,
            secret=secret,
            keys=[k["name"] for k in key_list] if key_list is not None else None,
            reason=reason,
            vault_name=vault_name,
            kv_version=kv_version,
        )._asdict()

    def _supply_key_values(keys: List[VaultKey]) -> Dict[str, str]:
        key_count = len(keys)
        new_values = {}
        percentage = 0
        increment = int(100 / key_count)

        for i, key in enumerate(keys, 1):
            key_name = key["name"]
            ui.output_progress(f"Waiting for value {i} of {key_count}...", percent=percentage, element_id="supply_new_values")
            if key["format"] == "string":
                if not key["sensitive"]:
                    value = ui.input_text(f"Enter the {secret} '{key_name}'", required=True)
                else:
                    while True:
                        value = ui.input_password(
                            f"Enter the {secret} '{key_name}' (sensitive)", element_id=f"new_value_{key_name}"
                        )
                        confirm_value = ui.input_password(
                            f"Confirm the {secret} '{key_name}' (sensitive)", element_id=f"new_value_{key_name}"
                        )
                        if value != confirm_value:
                            ui.output_error(
                                f"New '{key_name}' values do not match. Please try again.",
                                element_id=f"new_value_{key_name}_error",
                            )
                            continue
                        break
            elif key["format"] == "file":
                if not key["sensitive"]:
                    value = ui.input_file(
                        f"Upload a file with the secret data for '{key_name}'", element_id=f"new_value_{key_name}"
                    ).data
                else:
                    value = ui.input_sensitive_file(
                        f"Upload a file with the secret data for '{key_name}' (sensitive)", element_id=f"new_value_{key_name}"
                    ).data
            else:
                # The rogue_keys check at the start of this function *should* mean this never happens, but better safe than sorry.
                logger.error(f"{processed_keys=}")
                raise KeyError(
                    f"Supplied key '{key}' is not recognised as a sensitive/insensitive string or file. "
                    "Please view the logs for more information."
                )

            new_values[key_name] = value
            percentage = int(percentage + increment)

        ui.output_progress("All values for the secret have been supplied", percent=100, element_id="supply_new_values")
        return new_values

    def _create_missing_engine() -> None:
        ui.output_markdown(
            _markdown_templates.missing_engine.format(
                icon=_markdown_templates.icon,
                engine=engine,
                kv_version=kv_version,
                secret=secret,
                keys=_format_keys(processed_keys),
            )
        )
        if not ui.input_confirm("Do you want to create the engine & populate the secret?", confirm="Yes", cancel="No"):
            ui.output_error("Please ask an administrator to add the missing Vault configuration.")
            raise exceptions.InvalidEngine(f"No v{kv_version} secrets engine found with name '{engine}'.")
        kv_engine.create_engine(
            engine=engine, reason=f"Task is creating the secrets engine '{engine}'", vault_name=vault_name, kv_version=kv_version
        )
        # Now move on to creating the secret in the new engine
        _create_missing_secret(engine_created=True)

    def _create_missing_secret(engine_created: bool = False) -> None:
        if not engine_created:
            # Don't show this stuff if the user has already passed through _create_engine.
            ui.output_markdown(
                _markdown_templates.missing_secret.format(
                    icon=_markdown_templates.icon, engine=engine, secret=secret, keys=_format_keys(processed_keys)
                )
            )

            if not ui.input_confirm("Do you know the values for these keys?", confirm="Yes", cancel="No"):
                ui.output_error("Please ask an administrator to add the missing Vault configuration.")
                raise exceptions.InvalidSecret(f"Required secret '{secret}' does not exist in engine '{engine}'.")

        create_secret(
            secret=secret,
            data=_supply_key_values(processed_keys),
            engine=engine,
            reason=f"Task is creating required secret '{secret}'",
            vault_name=vault_name,
            kv_version=kv_version,
        )

    def _create_missing_keys(original_exception: Exception) -> None:
        # If this exception was raised the engine & secret already
        # existed, but one of more of the supplied keys didn't exist.
        # This function takes the original exception because it contains
        # specific information about which keys were missing.

        ui.output_progress("Getting missing values...", percent=None, element_id="get_missing_values")
        try:
            existing_secrets = _read(
                reason=f"Finding values missing from secret '{secret}'", key_list=None
            )  # Supply no keys as we want the whole secret
        except exceptions.EmptySecret as e:
            if "exists but contains no data" not in str(e):
                raise e
            # The secret exists but is empty, so all the keys are missing.
            existing_secrets = {}

        missing_keys = [key for key in processed_keys if key["name"] not in existing_secrets]
        missing_key_count = len(missing_keys)
        ui.output_progress(f"Found {missing_key_count} missing values", percent=100, element_id="get_missing_values")

        ui.output_markdown(
            _markdown_templates.missing_secret.format(
                icon=_markdown_templates.icon, keys=_format_keys(missing_keys), secret=secret
            )
        )

        if not ui.input_confirm("Do you know the values for the secret?", confirm="Yes", cancel="No"):
            ui.output_error("Please ask an administrator to add the missing secret values above.")
            raise original_exception

        update_secret(
            secret=secret,
            data=_supply_key_values(missing_keys),
            engine=engine,
            reason=f"Task is adding {missing_key_count} values to secret '{secret}'",
            vault_name=vault_name,
            kv_version=kv_version,
        )

    try:
        secrets = _read(reason=reason or f"Task is reading {len(processed_keys)} keys from secret '{secret}' in Hashicorp Vault")
    except (exceptions.InvalidEngine, exceptions.InvalidSecret, exceptions.InvalidKey, exceptions.EmptySecret,) as e:

        if not create_missing:
            # If create_missing is False, let it fail.
            raise e

        _missing_keys_handler = functools.partial(_create_missing_keys, e)

        try:
            {
                exceptions.InvalidEngine: _create_missing_engine,
                exceptions.InvalidSecret: _create_missing_secret,
                exceptions.EmptySecret: _missing_keys_handler,
                exceptions.InvalidKey: _missing_keys_handler,
            }[type(e)]()

            # Re-read the secrets after handling the error cases.
            secrets = _read(reason=f"Task is re-reading secret '{secret}' after modifying the vault")
        except KeyError as e:
            logger.exception(e)
            raise exceptions.VaultError(
                "Vault setup wizard encountered an unhandled exception. View the logs for more information."
            )

    return secrets
